@@ -1,15 +1,16 @@
+import json
 import sys
 from abc import ABC, abstractmethod
 from collections import defaultdict
 from enum import Enum
-from typing import Dict, Iterable, Iterator, Optional, Set, TextIO, Tuple, Type
+from typing import Dict, Iterable, Iterator, Optional, Set, Tuple, Type
 
 if sys.version_info < (3, 9):
     from astunparse import unparse
 else:
     from ast import unparse
 
-from .pickle import Interpreter, List, Pickled, Proto
+from fickling.fickle import Interpreter, List, Pickled, Proto
 
 
 class AnalyzerMeta(type):
@@ -70,11 +71,10 @@ class Analyzer(metaclass=AnalyzerMeta):
 
 class Severity(Enum):
     LIKELY_SAFE = (0, "No Unsafe Operations Discovered")
-    UNKNOWN = (1, "Unknown")
-    SUSPICIOUS = (2, "Suspicious")
-    LIKELY_UNSAFE = (3, "Likely Unsafe")
-    LIKELY_OVERTLY_MALICIOUS = (4, "Likely Overtly Malicious")
-    OVERTLY_MALICIOUS = (5, "Overtly Malicious")
+    SUSPICIOUS = (1, "Suspicious")
+    LIKELY_UNSAFE = (2, "Likely Unsafe")
+    LIKELY_OVERTLY_MALICIOUS = (3, "Likely Overtly Malicious")
+    OVERTLY_MALICIOUS = (4, "Overtly Malicious")
 
     def __lt__(self, other):
         return isinstance(other, Severity) and self.value < other.value
@@ -97,9 +97,19 @@ class Severity(Enum):
 
 
 class AnalysisResult:
-    def __init__(self, severity: Severity, message: Optional[str] = None):
+    def __init__(
+        self,
+        severity: Severity,
+        message: Optional[str] = None,
+        analysis_name: str = None,
+        trigger: Optional[str] = None,
+    ):
         self.severity: Severity = severity
         self.message: Optional[str] = message
+        self.analysis_name: str = analysis_name
+        self.trigger: Optional[str] = (
+            trigger  # Field to store the trigger code fragment or artifact
+        )
 
     def __lt__(self, other):
         return isinstance(other, AnalysisResult) and (
@@ -113,7 +123,7 @@ class AnalysisResult:
 
     def __str__(self):
         if self.message is None:
-            return ""
+            return "No issues found"
         else:
             return self.message
 
@@ -129,26 +139,21 @@ class Analysis(ABC):
         raise NotImplementedError()
 
 
-class ProtoAnalysis(Analysis):
+class DuplicateProtoAnalysis(Analysis):
     def analyze(self, context: AnalysisContext) -> Iterator[AnalysisResult]:
         had_proto = False
         proto_versions: Set[int] = set()
         for i, opcode in enumerate(context.pickled):
             if isinstance(opcode, Proto):
                 if had_proto:
-                    if i == 2:
-                        suffix = "rd"
-                    elif i == 1:
-                        suffix = "nd"
-                    elif i == 0:
-                        suffix = "st"
-                    else:
-                        suffix = "th"
+                    suffix = self._get_suffix(i)
                     if opcode.version in proto_versions:
                         yield AnalysisResult(
                             Severity.LIKELY_UNSAFE,
                             f"The {i + 1}{suffix} opcode is a duplicate PROTO, which is unusual "
                             f"and may be indicative of a tampered pickle",
+                            "DuplicateProtoAnalysis",
+                            trigger=i + 1,
                         )
                     else:
                         yield AnalysisResult(
@@ -156,17 +161,31 @@ class ProtoAnalysis(Analysis):
                             f"The {i + 1}{suffix} opcode is a duplicate PROTO with a different "
                             f"version than reported in the previous PROTO opcode, which is almost "
                             f"certainly a sign of a tampered pickle",
+                            "DuplicateProtoAnalysis",
+                            trigger=i + 1,
                         )
                 else:
                     had_proto = True
+                proto_versions.add(opcode.version)
+
+    @staticmethod
+    def _get_suffix(index: int) -> str:
+        return {0: "st", 1: "nd", 2: "rd"}.get(index, "th")
+
+
+class MisplacedProtoAnalysis(Analysis):
+    def analyze(self, context: AnalysisContext) -> Iterator[AnalysisResult]:
+        for i, opcode in enumerate(context.pickled):
+            if isinstance(opcode, Proto):
                 if opcode.version >= 2 and i > 0:
                     yield AnalysisResult(
                         Severity.LIKELY_UNSAFE,
                         f"The protocol version is {opcode.version}, but the PROTO opcode is not "
                         f"the first opcode in the pickle, as required for versions 2 and later; "
                         f"this may be indicative of a tampered pickle",
+                        "MisplacedProtoAnalysis",
+                        trigger=opcode.version,
                     )
-                proto_versions.add(opcode.version)
 
 
 class NonStandardImports(Analysis):
@@ -179,6 +198,8 @@ class NonStandardImports(Analysis):
                     f"`{shortened}` imports a Python module that is not a part of "
                     "the standard library; this can execute arbitrary code and is "
                     "inherently unsafe",
+                    "NonStandardImports",
+                    trigger=shortened,
                 )
 
 
@@ -198,17 +219,23 @@ class OvertlyBadEvals(Analysis):
                 or shortened.startswith("exec(")
                 or shortened.startswith("compile(")
                 or shortened.startswith("open(")
+                or shortened.startswith("_run_code(")
+                or shortened.startswith("execWrapper(")
             ):
                 # this is overtly bad, so record it and print it at the end
                 yield AnalysisResult(
                     Severity.OVERTLY_MALICIOUS,
                     f"Call to `{shortened}` is almost certainly evidence of a "
                     "malicious pickle file",
+                    "OvertlyBadEval",
+                    trigger=shortened,
                 )
             elif not already_reported:
                 yield AnalysisResult(
                     Severity.LIKELY_UNSAFE,
                     f"Call to `{shortened}` can execute arbitrary code and is inherently unsafe",
+                    "OvertlyBadEval",
+                    trigger=shortened,
                 )
 
 
@@ -219,6 +246,8 @@ class UnsafeImports(Analysis):
             yield AnalysisResult(
                 Severity.LIKELY_OVERTLY_MALICIOUS,
                 f"`{shortened}` is suspicious and indicative of an overtly malicious pickle file",
+                "UnsafeImports",
+                trigger=shortened,
             )
 
 
@@ -231,6 +260,8 @@ class UnusedVariables(Analysis):
                 Severity.SUSPICIOUS,
                 f"Variable `{varname}` is assigned value `{shortened}` but unused afterward; "
                 f"this is suspicious and indicative of a malicious pickle file",
+                "UnusedVariables",
+                trigger=(varname, shortened),
             )
 
 
@@ -242,42 +273,59 @@ class AnalysisResults:
     @property
     def severity(self) -> Severity:
         if not self.results:
-            return Severity.UNKNOWN
+            return Severity.LIKELY_SAFE
         return max(r.severity for r in self.results)
 
     def __bool__(self):
         """Returns True if all analyses failed to find any unsafe operations"""
         return all(map(bool, sorted(self.results)))
 
+    def detailed_results(self) -> Dict[str, Dict[str, str]]:
+        detailed = defaultdict(dict)
+        for result in self.results:
+            if result.trigger:
+                detailed["AnalysisResult"][result.analysis_name] = result.trigger
+        return dict(detailed)
+
     def to_string(self, verbosity: Severity = Severity.SUSPICIOUS):
         return "\n".join(str(r) for r in self.results if verbosity <= r.severity)
 
     __str__ = to_string
 
+    def to_dict(self, verbosity: Severity = Severity.SUSPICIOUS):
+        analysis_message = self.to_string(verbosity)
+        severity_data = {
+            "severity": self.severity.name,
+            "analysis": (
+                analysis_message
+                if analysis_message.strip()
+                else "Warning: Fickling failed to detect any overtly unsafe code,"
+                "but the pickle file may still be unsafe."
+                "Do not unpickle this file if it is from an untrusted source!\n\n"
+            ),
+            "detailed_results": self.detailed_results(),
+        }
+        return severity_data
+
 
 def check_safety(
     pickled: Pickled,
-    stdout: Optional[TextIO] = None,
-    stderr: Optional[TextIO] = None,
     analyzer: Optional[Analyzer] = None,
     verbosity: Severity = Severity.SUSPICIOUS,
+    json_output_path: Optional[str] = None,
 ) -> AnalysisResults:
-    if stdout is None:
-        stdout = sys.stdout
-    if stderr is None:
-        stderr = sys.stderr
-
     if analyzer is None:
         analyzer = Analyzer.default_instance
 
     results = analyzer.analyze(pickled)
-
-    stdout.write(results.to_string(verbosity))
-
-    if results.severity >= Severity.LIKELY_SAFE:
-        stderr.write(
-            "Warning: Fickling failed to detect any overtly unsafe code, but the pickle file may "
-            "still be unsafe.\n\nDo not unpickle this file if it is from an untrusted source!\n"
-        )
-
+    severity_data = results.to_dict(verbosity)
+    if json_output_path:
+        # This is intentionally "a" to handle the case of stacked pickles
+        with open(json_output_path, "a") as json_file:
+            json.dump(severity_data, json_file, indent=4)
     return results
+
+
+def is_likely_safe(filepath: str):
+    with open(filepath, "rb") as f:
+        return check_safety(Pickled.load(f)).severity == Severity.LIKELY_SAFE
