@@ -7,8 +7,10 @@ if sys.version_info >= (3, 9):
 else:
     from astunparse import unparse
 
-from . import __version__, pickle, tracing
-from .analysis import check_safety
+from . import __version__, fickle, tracing
+from .analysis import Severity, check_safety
+
+DEFAULT_JSON_OUTPUT_FILE = "safety_results.json"
 
 
 def main(argv: Optional[List[str]] = None) -> int:
@@ -16,7 +18,7 @@ def main(argv: Optional[List[str]] = None) -> int:
         argv = sys.argv
 
     parser = ArgumentParser(
-        description="fickle is a static analyzer and interpreter for Python pickle data"
+        description="fickling is a static analyzer and interpreter for Python pickle data"
     )
     parser.add_argument(
         "PICKLE_FILE",
@@ -33,7 +35,7 @@ def main(argv: Optional[List[str]] = None) -> int:
         "-i",
         type=str,
         default=None,
-        help="inject the specified Python code to be run at the end of depickling, "
+        help="inject the specified Python code to be run at the end of unpickling, "
         "and output the resulting pickle data",
     )
     parser.add_argument(
@@ -74,6 +76,21 @@ def main(argv: Optional[List[str]] = None) -> int:
             "even if this check exits with code zero."
         ),
     )
+
+    parser.add_argument(
+        "--json-output",
+        type=str,
+        default=None,
+        help="path to the output JSON file to store the analysis results from check-safety."
+        f"If not provided, a default file named {DEFAULT_JSON_OUTPUT_FILE} will be used.",
+    )
+
+    parser.add_argument(
+        "--print-results",
+        action="store_true",
+        help="Print the analysis results to the console when checking safety.",
+    )
+
     parser.add_argument(
         "--trace",
         "-t",
@@ -100,8 +117,8 @@ def main(argv: Optional[List[str]] = None) -> int:
         else:
             file = open(args.PICKLE_FILE, "rb")
         try:
-            stacked_pickled = pickle.StackedPickle.load(file)
-        except pickle.PickleDecodeError as e:
+            stacked_pickled = fickle.StackedPickle.load(file)
+        except fickle.PickleDecodeError as e:
             sys.stderr.write(f"Error: {str(e)}\n")
             return 1
         finally:
@@ -121,7 +138,7 @@ def main(argv: Optional[List[str]] = None) -> int:
             for pickled in stacked_pickled[: args.inject_target]:
                 pickled.dump(buffer)
             pickled = stacked_pickled[args.inject_target]
-            if not isinstance(pickled[-1], pickle.Stop):
+            if not isinstance(pickled[-1], fickle.Stop):
                 sys.stderr.write(
                     "Warning: The last opcode of the input file was expected to be STOP, but was "
                     f"in fact {pickled[-1].info.name}"
@@ -136,14 +153,28 @@ def main(argv: Optional[List[str]] = None) -> int:
                 pickled.dump(buffer)
         elif args.check_safety:
             was_safe = True
+            json_output_path = args.json_output or DEFAULT_JSON_OUTPUT_FILE
             for pickled in stacked_pickled:
-                if not check_safety(pickled):
+                safety_results = check_safety(pickled, json_output_path=json_output_path)
+
+                # Print results if requested
+                if args.print_results:
+                    print(safety_results.to_string())
+
+                if safety_results.severity > Severity.LIKELY_SAFE:
                     was_safe = False
+                    if args.print_results:
+                        sys.stderr.write(
+                            "Warning: Fickling detected that the pickle file may be unsafe.\n\n"
+                            "Do not unpickle this file if it is from an untrusted source!\n\n"
+                        )
+
             return [1, 0][was_safe]
+
         else:
             var_id = 0
             for i, pickled in enumerate(stacked_pickled):
-                interpreter = pickle.Interpreter(
+                interpreter = fickle.Interpreter(
                     pickled, first_variable_id=var_id, result_variable=f"result{i}"
                 )
                 if args.trace:
@@ -153,14 +184,14 @@ def main(argv: Optional[List[str]] = None) -> int:
                     print(unparse(interpreter.to_ast()))
                 var_id = interpreter.next_variable_id
     else:
-        pickled = pickle.Pickled(
+        pickled = fickle.Pickled(
             [
-                pickle.Global.create("__builtin__", "eval"),
-                pickle.Mark(),
-                pickle.Unicode(args.create.encode("utf-8")),
-                pickle.Tuple(),
-                pickle.Reduce(),
-                pickle.Stop(),
+                fickle.Global.create("__builtin__", "eval"),
+                fickle.Mark(),
+                fickle.Unicode(args.create.encode("utf-8")),
+                fickle.Tuple(),
+                fickle.Reduce(),
+                fickle.Stop(),
             ]
         )
         if args.PICKLE_FILE == "-":
